/*
 *	This file is to be linked with the rest of the libraries in
 *	the distribution package.
 *
 * $Id: ptmodule.c,v 1.5 2005/04/07 07:02:15 robert Exp $
 */

#include <linux/config.h>
#include <linux/version.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/types.h>
#include <linux/slab.h>
#include <linux/delay.h>
#if (LINUX_VERSION_CODE >= 131343) /* 2.1.15 */
#include <linux/init.h>
#endif

#include "include/pctel.h"
#include "include/halwraps.h"
#include "include/vuart.h"

static char* pctel_options __initdata =
#if defined (HAL_I8XX) || defined (HAL_VIA686A)
#if defined (INTEL)
	"INTEL i8XX"
#elif defined (VIA)
	"VIA686A"
#elif defined (SIS540)
        "SIS540"
#else
	"*UNKNOWN*"
#endif
#ifdef AUDIO_ROUTING
	" with AUDIO_ROUTING"
#endif
#elif defined (HAL_PCT789)
	"PCT789"
#elif defined (HAL_CM8738)
	"CM8738"
#else
	"*UNKNOWN*"
#endif
	;
									    

void Pctel_module_init(void)
{
	MOD_INC_USE_COUNT;
}

void Pctel_module_free(void)
{
	MOD_DEC_USE_COUNT;
}

int Pctel_module_is_free(void)
{
	return (MOD_IN_USE == 0);
}

void *pctel_kmalloc(size_t size, int flags)
{
	return kmalloc(size, GFP_KERNEL|GFP_ATOMIC);
}

void pctel_kfree(const void *objp)
{
	kfree(objp);
}

void pctel___udelay(unsigned long usecs)
{
	__udelay(usecs);
}

static int __init pctel_init(void)
{
	printk( KERN_INFO "PCTel hardware driver version %s for %s\n",
		VERSION, pctel_options );
	return 0;
}

static void __exit pctel_cleanup(void) 
{
	printk( KERN_INFO "Unloading PCTel hardware driver version %s\n",
		VERSION );
}


#define MODULE_NAME	"pctel"

module_init(pctel_init);
module_exit(pctel_cleanup);

MODULE_DESCRIPTION("PCtel modem driver");

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,10))
MODULE_LICENSE("GPL linked with proprietary libraries");
#endif
