/*
 *  linux/drivers/char/8250.h
 *
 *  Driver for 8250/16550-type serial ports
 *
 *  Based on drivers/char/serial.c, by Linus Torvalds, Theodore Ts'o.
 *
 *  Copyright (C) 2001 Russell King.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 *  $Id: ptserial-2.6.h,v 1.7 2007/03/10 17:21:04 robert Exp $
 *
 * 04/05: Added Intel i855PN chipset, ICH4/4M AC'97 Controller (8086:24c6)
 *	      Robert Thornburrow <robert.thornburrow@tsac.fsnet.co.uk>
 * 07/05: Fix by Jeff Trull to allow compilation on kernels prior to 2.6.9.
 */

#ifndef __PTSERIAL_2_6_H
#define __PTSERIAL_2_6_H

/* Fix for removed config.h in 2.6.18 - 03/01/2007 */
#include <linux/version.h>

#if ( LINUX_VERSION_CODE < KERNEL_VERSION(2,6,18) )
#include <linux/config.h>
#endif

/* Fix by Jeff Trull for Debian Sarge (Kernel 2.6.8) - 20/07/2005
 * __iomem is used in kernels 2.6.9 onwards, so need to remove it */
#if ( LINUX_VERSION_CODE < KERNEL_VERSION(2,6,9) )
#ifndef __iomem
#define __iomem
#endif /* __iomem */
#endif /* kernel 2.6.9 */

#define UART_CAP_FIFO	(1 << 8)	/* UART has FIFO */
#define UART_CAP_EFR	(1 << 9)	/* UART has EFR */
#define UART_CAP_SLEEP	(1 << 10)	/* UART has IER sleep */
#define UART_CAP_AFE	(1 << 11)	/* MCR-based hw flow control */

#undef SERIAL_DEBUG_PCI

#define PROBE_RSA	(1 << 0)
#define PROBE_ANY	(~0)

#define HIGH_BITS_OFFSET ((sizeof(long)-sizeof(int))*8)


/* serial_core.h has no include protection, so fix */
#ifndef UPF_FOURPORT
# include <linux/serial_core.h>
#endif
#include <linux/device.h>

struct plat_ptserial_port {
	unsigned long	iobase;		/* io base address */
	void __iomem	*membase;	/* ioremap cookie or NULL */
	unsigned long	mapbase;	/* resource base */
	unsigned int	irq;		/* interrupt number */
	unsigned int	uartclk;	/* UART clock rate */
	unsigned char	regshift;	/* register shift */
	unsigned char	iotype;		/* UPIO_* */
	unsigned int	flags;		/* UPF_* flags */
};

#include "include/pctel.h"
#include "include/vuart.h"
#include "include/halwraps.h"

/* uart_port doesn't contain region_0, region_1, so we map to iobase, membase */
#define pctel_region_0(x)    (x).iobase
#define pctel_region_0_p(x)  (x)->iobase
#define pctel_region_1(x)    (x).membase
#define pctel_region_1_p(x)  (x)->membase

#define PCTEL_FLAGS (ASYNC_BOOT_AUTOCONF | ASYNC_SKIP_TEST)

/* override supported serial type definition in serial.h */
#define PORT_PCTEL                      14

#ifdef  PORT_MAX
#undef  PORT_MAX
#endif
#define PORT_MAX                        14

#define PCTEL_MAJOR 62
#define PCTEL_MINOR 64

#define PCTEL_TTY_LINE			0



/* VENDOR/DEVICEID List */
/*  PCTEL ASIC (PCT789T) */
#define PCI_VENDOR_ID_PCTEL		0x134D
#define PCI_DEVICE_ID_PCTEL_PCT7890	0x7890
#define PCI_DEVICE_ID_PCTEL_PCT7891	0x7891
#define PCI_DEVICE_ID_PCTEL_PCT7892	0x7892
#define PCI_DEVICE_ID_PCTEL_PCT7893	0x7893
#define PCI_DEVICE_ID_PCTEL_PCT7894	0x7894
#define PCI_DEVICE_ID_PCTEL_PCT7895	0x7895
#define PCI_DEVICE_ID_PCTEL_PCT7896	0x7896
#define PCI_DEVICE_ID_PCTEL_PCT7897	0x7897

/*  C-MEDIA ASIC (CM8738) */
#ifdef PCI_VENDOR_ID_CMEDIA /* Get rid of warning */
#undef PCI_VENDOR_ID_CMEDIA
#endif
#define PCI_VENDOR_ID_CMEDIA		0x13F6
#ifdef PCI_DEVICE_ID_CMEDIA_CM8738 /* Get rid of warning */
#undef PCI_DEVICE_ID_CMEDIA_CM8738
#endif
#define PCI_DEVICE_ID_CMEDIA_CM8738	0x211

/*  INTEL i8xx MR */
#define PCI_VENDOR_ID_INTEL		0x8086
#define PCI_DEVICE_ID_I810_MR		0x2416   /*  desktop */
#define PCI_DEVICE_ID_I810_MR_LAPTOP    0x7196   /*  laptop  */
#define PCI_DEVICE_ID_I820		0x2446
#define PCI_DEVICE_ID_I845		0x2486
#define PCI_DEVICE_ID_I855		0x24c6   /* i855PN laptop */

/*  SIS MR */
#define PCI_VENDOR_ID_SIS               0x1039 
#define PCI_DEVICE_ID_SIS540_MR         0x7013

/*  VIA MR */
#define PCI_VENDOR_ID_VIA		0x1106
#define PCI_DEVICE_ID_VIA686A_MR	0x3068

/* END VENDOR/DEVICEID List */

#ifdef AUDIO_ROUTING
extern AUDIO_BUFFERS audio_buffers;
#endif


extern int pctel_new(struct uart_port *port);
extern int pctel_init(void);
extern void pctel_exit(void);

extern int ptserial_hw_check_modem(int vendor);

#endif /* __PTSERIAL_2_6_H */
