/*
 *      This file is to be linked with the rest of the libraries in
 *      the distribution package.  It links to proprietry code and
 *      is therefore NOT covered by the GPL.
 *
 * $Id: ptserial_hw-2.6.c,v 1.4 2007/03/10 17:21:04 robert Exp $
 */
#include <linux/version.h>

/* Fix for removed config.h in kernel 2.6.18 - 03/01/2007 */
#if ( LINUX_VERSION_CODE < KERNEL_VERSION(2,6,18) )
#include <linux/config.h>
#endif

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/types.h>
#include <linux/slab.h>
#include <linux/delay.h>
#include <linux/init.h>

#include "include/pctel.h"
#include "include/halwraps.h"
#include "include/vuart.h"

#define PCI_VENDOR_ID_PCTEL     0x134D
#define PCI_VENDOR_ID_CMEDIA    0x13F6
#define PCI_VENDOR_ID_INTEL     0x8086
#define PCI_VENDOR_ID_SIS       0x1039 
#define PCI_VENDOR_ID_VIA       0x1106



#define MY_NAME "pctel_hw"

#define err(format, arg...) \
	printk(KERN_ERR "%s(%d): " format "\n", MY_NAME, __LINE__, ## arg)
#define info(format, arg...) \
	printk(KERN_INFO "%s(%d): " format "\n", MY_NAME, __LINE__, ## arg)
#define warn(format, arg...) \
	printk(KERN_WARNING "%s(%d): " format "\n", MY_NAME, __LINE__, ## arg)


static char* pctel_options __initdata =
#if defined (HAL_I8XX) || defined (HAL_VIA686A)
#if defined (INTEL)
        "INTEL i8xx"
#elif defined (VIA)
        "VIA686A"
#elif defined (SIS540)
        "SIS540"
#else
        "*UNKNOWN*"
#endif
#ifdef AUDIO_ROUTING
        " with AUDIO_ROUTING"
#endif
#elif defined (HAL_PCT789)
        "PCT789"
#elif defined (HAL_CM8738)
        "CM8738"
#else
        "*UNKNOWN*"
#endif
        ;

/**
 * This function is necessary in order to enforce a dependency between pctel.ko
 * and pctel_hw.ko.  It provides a quick check to see that the correct HAL
 * macro has been defined during compilation giving some idea whether we are
 * linked against the correct binary code.
 */
int ptserial_hw_check_modem( int vendor )
{
    int ret = 0;
#if defined (HAL_PCT789)
    ret = (vendor == PCI_VENDOR_ID_PCTEL);
#elif defined (HAL_CM8738)
    ret = (vendor == PCI_VENDOR_ID_CMEDIA);
#elif defined (HAL_I8XX) || defined (HAL_VIA686A)
#if defined (INTEL)
    ret = (vendor == PCI_VENDOR_ID_INTEL);
#elif defined (SIS540)
    ret = (vendor == PCI_VENDOR_ID_SIS);
#elif defined (VIA)
    ret = (vendor == PCI_VENDOR_ID_VIA);
#endif
#endif
    
    return ret;
}

EXPORT_SYMBOL(ptserial_hw_check_modem);

/**
 * Symbols defined in binary.a, grouped by contributing object file.
 */

/* HAL */
EXPORT_SYMBOL(HAL_Deinit);
EXPORT_SYMBOL(HAL_DoInterrupt);
EXPORT_SYMBOL(HAL_GetTimer);
EXPORT_SYMBOL(HAL_Init);
 
/* Controller */
EXPORT_SYMBOL(GlobalTimer);
EXPORT_SYMBOL(PctelInitCtrlVars);
EXPORT_SYMBOL(modem_main);

/* DSP */
EXPORT_SYMBOL(dspMain);

/* vuart */
EXPORT_SYMBOL(COM_Vlcr);
EXPORT_SYMBOL(PctelInitVUartVars);
EXPORT_SYMBOL(get_uart_dlm);
EXPORT_SYMBOL(get_uart_dll);
EXPORT_SYMBOL(get_uart_ier);
EXPORT_SYMBOL(get_uart_iir);
EXPORT_SYMBOL(get_uart_lsr);
EXPORT_SYMBOL(get_uart_lcr);
EXPORT_SYMBOL(get_uart_mcr);
EXPORT_SYMBOL(get_uart_msr);
EXPORT_SYMBOL(get_uart_rx);
EXPORT_SYMBOL(get_uart_scr);
EXPORT_SYMBOL(put_uart_dll);
EXPORT_SYMBOL(put_uart_dlm);
EXPORT_SYMBOL(put_uart_ier);
EXPORT_SYMBOL(put_uart_iir);
EXPORT_SYMBOL(put_uart_lcr);
EXPORT_SYMBOL(put_uart_lsr);
EXPORT_SYMBOL(put_uart_mcr);
EXPORT_SYMBOL(put_uart_msr);
EXPORT_SYMBOL(put_uart_scr);
EXPORT_SYMBOL(put_uart_tx);

#ifdef AUDIO_ROUTING
EXPORT_SYMBOL(audio_buffers);
EXPORT_SYMBOL(audio_on_count);
EXPORT_SYMBOL(ccount);
EXPORT_SYMBOL(dsp_buffer);
EXPORT_SYMBOL(dsp_head);
EXPORT_SYMBOL(dsp_tail);
#endif /* AUDIO_ROUTING */


/**
 * Module init function.  Doesn't need to do anything, just reports a string
 * indicating the driver version with the given compile options has been
 * loaded.
 */
static int __init ptserial_hw_init( void )
{
        info( "PCTel hardware driver version %s for %s\n",
              VERSION, pctel_options );
        return 0;
}

/**
 * Module exit function.  Doesn't need to do anything, just reports the driver
 * version has been unloaded.
 */
static void __exit ptserial_hw_exit( void )
{
        info( "Unloading PCTel hardware driver version %s\n", VERSION );
}

module_init(ptserial_hw_init);
module_exit(ptserial_hw_exit);

MODULE_LICENSE("GPL linked with proprietary libraries");
MODULE_DESCRIPTION("PCTEL hardware modem driver: binary part of driver");
